/*
 * Bluetooth Controlled 8x8 RGB LED Matrix
 *
 * HSV to rgb24 Color Conversion and Generation Functions
 *
 * Written by: Craig A. Lindley
 * Last Update: 11/11/2014
 */

/*******************************************************************/
/***           Color Constants, Variables and Functions          ***/
/*******************************************************************/

#define NUM_OF_COLOR_VALUES 256
#define MIN_COLOR_VALUE       0
#define MAX_COLOR_VALUE     255

// HSV to RGB color conversion
// Input arguments
// hue in degrees (0 - 360.0)
// saturation (0.0 - 1.0)
// value (0.0 - 1.0)
// Output arguments
// red, green blue (0.0 - 1.0)
void hsvToRGB(float hue, float saturation, float value, float * red, float * green, float * blue) {

    int i;
    float f, p, q, t;

    if (saturation == 0) {
        // achromatic (grey)
        *red = *green = *blue = value;
        return;
    }
    hue /= 60;                  // sector 0 to 5
    i = floor(hue);
    f = hue - i;                // factorial part of h
    p = value * (1 - saturation);
    q = value * (1 - saturation * f);
    t = value * (1 - saturation * (1 - f));
    switch (i) {
    case 0:
        *red = value;
        *green = t;
        *blue = p;
        break;
    case 1:
        *red = q;
        *green = value;
        *blue = p;
        break;
    case 2:
        *red = p;
        *green = value;
        *blue = t;
        break;
    case 3:
        *red = p;
        *green = q;
        *blue = value;
        break;
    case 4:
        *red = t;
        *green = p;
        *blue = value;
        break;
    default:
        *red = value;
        *green = p;
        *blue = q;
        break;
    }
}

// Create a HSV color
rgb24 createHSVColor(float hue, float saturation,  float value) {

    float r, g, b;
    rgb24 color;

    hsvToRGB(hue, saturation, value, &r, &g, &b);

    color.red   = r * MAX_COLOR_VALUE;
    color.green = g * MAX_COLOR_VALUE;
    color.blue =  b * MAX_COLOR_VALUE;

    return color;
}

// Create an HSV color
rgb24 createHSVColor(int divisions, int index, float saturation, float value) {

    float hueAngle = (360.0 * index) / divisions;
    return createHSVColor(hueAngle, saturation, value);
}

// Create an HSV color
rgb24 createAHSVColor(int divisions, int index, float saturation, float value) {

    index %= divisions;
    return createHSVColor(divisions, index, saturation, value);
}

rgb24 createHSVColorWithDivisions(int divisions, int index) {

    return createAHSVColor(divisions, index, 1.0, 1.0);
}

// Input a value 0 to 255 to get a color value.
// The colors are a transition r - g - b - back to r.
rgb24 colorWheel(int wheelPos) {

    rgb24 color;

    wheelPos %= 256;

    if (wheelPos < 85) {
        color.red   = wheelPos * 3;
        color.green = 255 - wheelPos * 3;
        color.blue  = 0;
    }    
    else if (wheelPos < 170) {
        wheelPos -= 85;
        color.red   = 255 - wheelPos * 3;
        color.green = 0;
        color.blue  = wheelPos * 3;
    } 
    else    {
        wheelPos -= 170;
        color.red   = 0;
        color.green = wheelPos * 3;
        color.blue  = 255 - wheelPos * 3;
    }
    return color;
}


