/*
 * Bluetooth Controlled 8x8 RGB LED Matrix
 *
 * Driver code for the LED matrix
 *
 * Written by: Craig A. Lindley
 * Last Update: 11/11/2014
 */

#include "spi4teensy3.h"
#include "TimerOne.h"

#define USE_GAMMA_CORRECTION 1

#define MATRIX_WIDTH  8
#define MATRIX_HEIGHT 8

// PWM driver control pins
#define LATCH_PIN  9
#define BLANK_PIN 10

// LED Matrix row select pins
#define ROW0_PIN  16
#define ROW1_PIN  17
#define ROW2_PIN  18
#define ROW3_PIN  19
#define ROW4_PIN  20
#define ROW5_PIN  21
#define ROW6_PIN  22
#define ROW7_PIN  23

// Interrupt period calculations
#define DATA_XFER_TIME_USEC     13
#define ROW_DISPLAY_TIME_USEC 1024

#define INTERRUPT_TIME_USEC (DATA_XFER_TIME_USEC + ROW_DISPLAY_TIME_USEC)

// Frame buffer definition
#define NUMBER_OF_ROWS 8
#define BYTES_PER_ROW 36

#if (USE_GAMMA_CORRECTION == 1) 

// Gamma correction table
const byte GAMMA_TABLE [] = {
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  1,  1,  1,
    1,  1,  1,  1,  1,  1,  1,  1,  1,  2,  2,  2,  2,  2,  2,  2,
    2,  3,  3,  3,  3,  3,  3,  3,  4,  4,  4,  4,  4,  5,  5,  5,
    5,  6,  6,  6,  6,  7,  7,  7,  7,  8,  8,  8,  9,  9,  9, 10,
    10, 10, 11, 11, 11, 12, 12, 13, 13, 13, 14, 14, 15, 15, 16, 16,
    17, 17, 18, 18, 19, 19, 20, 20, 21, 21, 22, 22, 23, 24, 24, 25,
    25, 26, 27, 27, 28, 29, 29, 30, 31, 32, 32, 33, 34, 35, 35, 36,
    37, 38, 39, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 50,
    51, 52, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 66, 67, 68,
    69, 70, 72, 73, 74, 75, 77, 78, 79, 81, 82, 83, 85, 86, 87, 89,
    90, 92, 93, 95, 96, 98, 99,101,102,104,105,107,109,110,112,114,
    115,117,119,120,122,124,126,127,129,131,133,135,137,138,140,142,
    144,146,148,150,152,154,156,158,160,162,164,167,169,171,173,175,
    177,180,182,184,186,189,191,193,196,198,200,203,205,208,210,213,
    215,218,220,223,225,228,231,233,236,239,241,244,247,249,252,255
};

#endif

// Frame buffer is 2D array of bytes
byte frameBuffer[NUMBER_OF_ROWS][BYTES_PER_ROW];

/*******************************************************************/
/***                  Interrupt Service Routine                  ***/
/*******************************************************************/

void rowInterrupt(void) {

    static byte rowNumber = 0;

    // Get pointer to row data for this interrupt in frameBuffer
    byte *ptr = &frameBuffer[rowNumber][0];

    // Send the row data to the TLC5947
    spi4teensy3::send(ptr, BYTES_PER_ROW);

    // Set BLANK high to disable PWM outputs and reset grayscale clock
    digitalWrite(BLANK_PIN, HIGH);

    // Strobe to latch row data into the chip
    digitalWrite(LATCH_PIN, HIGH);
    digitalWrite(LATCH_PIN, LOW);

    // Unselect the previous row (set row pin to input) then select new row (set row pin to output). 
    switch(rowNumber) {
    case 0:
        pinMode(ROW7_PIN, INPUT);
        pinMode(ROW0_PIN, OUTPUT);
        break;

    case 1:
        pinMode(ROW0_PIN, INPUT);
        pinMode(ROW1_PIN, OUTPUT);
        break;

    case 2:
        pinMode(ROW1_PIN, INPUT);
        pinMode(ROW2_PIN, OUTPUT);
        break;

    case 3:
        pinMode(ROW2_PIN, INPUT);
        pinMode(ROW3_PIN, OUTPUT);
        break;

    case 4:
        pinMode(ROW3_PIN, INPUT);
        pinMode(ROW4_PIN, OUTPUT);
        break;

    case 5:
        pinMode(ROW4_PIN, INPUT);
        pinMode(ROW5_PIN, OUTPUT);
        break;

    case 6:
        pinMode(ROW5_PIN, INPUT);
        pinMode(ROW6_PIN, OUTPUT);
        break;

    case 7:
        pinMode(ROW6_PIN, INPUT);
        pinMode(ROW7_PIN, OUTPUT);
        break;
    }
    // Delay for settling
    delayMicroseconds(5);

    // Set BLANK low to enable PWM outputs to drive LEDs
    // and starts the grayscale counter
    digitalWrite(BLANK_PIN, LOW);

    // Setup for next interrupt
    if (++rowNumber >= NUMBER_OF_ROWS) {
        rowNumber = 0;
    }
}

/*******************************************************************/
/***                    Frame Buffer Functions                   ***/
/*******************************************************************/

// Clear frame buffer memory
void clearFrameBuffer() {
    memset(frameBuffer, 0, sizeof(frameBuffer));
}

// Fill the frame buffer memory with specified color
void fillFrameBuffer(rgb24 color) {

    for (int row = 0; row < MATRIX_HEIGHT; row++) {
        for (int col = 0; col < MATRIX_WIDTH; col++) {
            setPixel(col, row, color);
        }
    }
}


// Set the 12 bit PWM value for the specified channel in frameBuffer
// row is 0..7; channelNumber is 0..23; pwmValue is 0..4095
// NOTE: teensy is little-endian and each PWM value requires 12 bits of storage
void setPWM12Value(uint8_t row, uint8_t channelNumber, uint16_t pwmValue) {

    byte ch = 23 - channelNumber;
    byte i = ch * 3 / 2;

    if ((ch % 2) == 0) {
        frameBuffer[row][i] = (pwmValue >> 4);
        i++;
        frameBuffer[row][i] = (frameBuffer[row][i] & 0x0F) | (byte) (pwmValue << 4);
    }  
    else  {
        frameBuffer[row][i] = (frameBuffer[row][i] & 0xF0) | (pwmValue >> 8);
        i++;
        frameBuffer[row][i] = (byte) pwmValue;
    }
}

// Get the 12 bit PWM value for the specified channel from frameBuffer.
// row is 0..7; channelNumber is 0..23; returned PWM value is 0..4095
uint16_t getPWM12Value(uint8_t row, uint8_t channelNumber) {

    uint16_t pwmValue;
    byte ch = 23 - channelNumber;
    byte i = ch * 3 / 2;

    if ((ch % 2) == 0) {
        pwmValue = frameBuffer[row][i] << 4;
        i++;
        pwmValue |= (frameBuffer[row][i] & 0xF0) >> 4;
    }  
    else  {
        pwmValue = (frameBuffer[row][i] & 0x0F) << 8;
        i++;
        pwmValue |= frameBuffer[row][i];
    }
    return pwmValue;
}

// Map 12 bit value to 8 bit value
uint8_t map12BitValueTo8BitValue(int value) {
    return (map(value, 0, 4095, 0, 255));
}

// Map 8 bit value to 12 bit value
uint16_t map8BitValueTo12BitValue(int value) {
    return (map(value, 0, 255, 0, 4095));
}

// Set the RGB pixel at (col, row) to the specified 36 bit color
// col is 0..7; row is 0..7
// 0, 0 is RGB pixel at the lower left
void setPixel(byte col, byte row, int redPWM, int greenPWM, int bluePWM) {

    setPWM12Value(row, col * 3,     redPWM);
    setPWM12Value(row, col * 3 + 1, greenPWM);
    setPWM12Value(row, col * 3 + 2, bluePWM);
}

// Set the RGB pixel at (col, row) to the specified 24 bit color
// col is 0..7; row is 0..7
// 0, 0 is RGB pixel at the lower left
void setPixel(byte col, byte row, rgb24 color) {
    
#if (USE_GAMMA_CORRECTION == 1) 
    color.red   = GAMMA_TABLE[color.red];
    color.green = GAMMA_TABLE[color.green];
    color.blue  = GAMMA_TABLE[color.blue];
#endif

    setPWM12Value(row, col * 3,     map8BitValueTo12BitValue(color.red));
    setPWM12Value(row, col * 3 + 1, map8BitValueTo12BitValue(color.green));
    setPWM12Value(row, col * 3 + 2, map8BitValueTo12BitValue(color.blue));
}

// Get the rgb24 color value of the RGB pixel at (col, row)
// col is 0..7; row is 0..7
// 0, 0 is RGB pixel at the lower left
rgb24 getPixel(byte col, byte row) {

    rgb24 color;

    uint16_t red   = getPWM12Value(row, col * 3);
    uint16_t green = getPWM12Value(row, col * 3 + 1);
    uint16_t blue  = getPWM12Value(row, col * 3 + 2);

    color.red   =   map12BitValueTo8BitValue(red);
    color.green =   map12BitValueTo8BitValue(green);
    color.blue  =   map12BitValueTo8BitValue(blue);

    return color;
}

// Set the RGB pixel by number to the specified 24 bit color
// pixelNumber is 0..63
// Pixel 0 is at the lower left
void setPixel(byte pixelNumber, rgb24 color) {
    int col = pixelNumber % 8;
    int row = pixelNumber / 8;
    setPixel(col, row, color);
}

/*******************************************************************/
/***               LED Matrix Driver Initialization              ***/
/*******************************************************************/

void initializeDriver() {

    // Setup the hardware
    pinMode(LATCH_PIN, OUTPUT);
    digitalWrite(LATCH_PIN, HIGH);

    pinMode(BLANK_PIN, OUTPUT);
    digitalWrite(BLANK_PIN, HIGH);

    // Set all row select pins to inputs and set low 
    // so that when changed to outputs they will be low.
    pinMode(ROW0_PIN, OUTPUT);
    digitalWrite(ROW0_PIN, LOW);
    pinMode(ROW0_PIN, INPUT);

    pinMode(ROW1_PIN, OUTPUT);
    digitalWrite(ROW1_PIN, LOW);
    pinMode(ROW1_PIN, INPUT);

    pinMode(ROW2_PIN, OUTPUT);
    digitalWrite(ROW2_PIN, LOW);
    pinMode(ROW2_PIN, INPUT);

    pinMode(ROW3_PIN, OUTPUT);
    digitalWrite(ROW3_PIN, LOW);
    pinMode(ROW3_PIN, INPUT);

    pinMode(ROW4_PIN, OUTPUT);
    digitalWrite(ROW4_PIN, LOW);
    pinMode(ROW4_PIN, INPUT);

    pinMode(ROW5_PIN, OUTPUT);
    digitalWrite(ROW5_PIN, LOW);
    pinMode(ROW5_PIN, INPUT);

    pinMode(ROW6_PIN, OUTPUT);
    digitalWrite(ROW6_PIN, LOW);
    pinMode(ROW6_PIN, INPUT);

    pinMode(ROW7_PIN, OUTPUT);
    digitalWrite(ROW7_PIN, LOW);
    pinMode(ROW7_PIN, INPUT);

    // Setup SPI interface for fastest speed
    spi4teensy3::init(); 

    // Clear frame buffer
    clearFrameBuffer();

    // Setup timer interrupt for multiplexing
    Timer1.initialize(INTERRUPT_TIME_USEC);
    Timer1.attachInterrupt(rowInterrupt);
}




