/*
 * Bluetooth Controlled 8x8 RGB LED Matrix
 *
 * Mode class definitions
 *
 * Written by: Craig A. Lindley
 * Last Update: 11/11/2014
 */

#include "BaseClass.h"
#include "Font.h"
#include "Patterns.h"
#include "Types.h"

// Fixup forward references
extern GLOBAL_PARAMETERS gParms;
extern void clearFrameBuffer();
extern void fillFrameBuffer(rgb24 color);
extern rgb24 createHSVColor(float hue, float saturation,  float value);
extern void setPixel(byte col, byte row, rgb24 color);

// Time patterns are displayed before changing
#define PATTERN_DISPLAY_TIME_SECONDS 50
#define PATTERN_DISPLAY_TIME_MS     (PATTERN_DISPLAY_TIME_SECONDS  * 1000)

#ifndef MODES_H
#define MODES_H

/*******************************************************************/
/***                       Device Off Mode                       ***/
/*******************************************************************/

class OffMode : 
public 
BaseClass {

public:

    OffMode() : 
    BaseClass(NULL) {
    }

    void initStateMachine() {

        // Clear the matrix
        clearFrameBuffer();
    }

    void runStateMachine(boolean updated) {
        // Do nothing except return
    }
};   

/*******************************************************************/
/***                       Mood Light Mode                       ***/
/*******************************************************************/

class MoodLightMode : 
public 
BaseClass {
    rgb24 color;

public:

    MoodLightMode(GLOBAL_PARAMETERS *_ptrParms) : 
    BaseClass(_ptrParms) {
    }

    void initStateMachine() {

        // Create color from global parameters
        color = createHSVColor(ptrParms->hue, ptrParms->sat,  ptrParms->val); 

        // Fill the matrix with the color
        fillFrameBuffer(color);
    }

    void runStateMachine(boolean updated) {
        if (updated) {
            // Create new color from global parameters
            color = createHSVColor(ptrParms->hue, ptrParms->sat,  ptrParms->val); 

            // Fill the matrix with the color
            fillFrameBuffer(color);
        }
    }
};   

/*******************************************************************/
/***                       Text Message Mode                     ***/
/*******************************************************************/

class TextMsgMode : 
public BaseClass {

    enum TMState {
        STATE_INIT, STATE_LENGTH_TEST, STATE_GET_CHAR, STATE_COLUMN_TEST, STATE_GET_COLUMN, STATE_DELAY
    };

    enum TMState state;

    int stringIndex;
    int stringLength;
    int delayTimeMS;
    int delayTime;

    int charIndex;
    int columnIndex;

    // Font data bitmap
    uint8_t fontData[8];

    // Shift the font bitmap to the left
    void shiftFontData() {

        fontData[0] = fontData[1];
        fontData[1] = fontData[2];
        fontData[2] = fontData[3];
        fontData[3] = fontData[4];
        fontData[4] = fontData[5];
        fontData[5] = fontData[6];
        fontData[6] = fontData[7];
    }

    // Render the font data bitmap to the display
    void renderFontData(rgb24 foregroundColor, rgb24 backgroundColor) {

        for (int col = 0; col < 8; col++) {
            uint8_t columnData = fontData[col];
            for (int bit = 0; bit < 8; bit++) {
                if (columnData & (1 << bit)) {
                    setPixel(col, 7 - bit, foregroundColor);
                }    
                else    {
                    setPixel(col, 7 - bit, backgroundColor);
                }
            }
        }
    }

public:

    TextMsgMode(GLOBAL_PARAMETERS *_ptrParms) : 
    BaseClass(_ptrParms) {
    }

    void initStateMachine() {

        // Calculate scrolling delay
        delayTimeMS = map(ptrParms->speed, 1, 10, 1000, 40);

        // Get text msg string length
        stringLength = strlen(ptrParms->textMsgBuffer);

        state = STATE_INIT;
    }

    void runStateMachine(boolean updated) {

        if (updated) {
            // Calculate scrolling delay
            delayTimeMS = map(ptrParms->speed, 1, 10, 1000, 40);
        }

        // Do until mode changed
        while (true) {
            switch (state) {

            case STATE_INIT:
                {
                    // Clear display
                    clearFrameBuffer();

                    // Clear font data
                    memset(fontData, 0, sizeof(fontData));

                    stringIndex = 0;

                    state = STATE_LENGTH_TEST;
                }
                break;

            case STATE_LENGTH_TEST:
                {
                    if (stringIndex >= stringLength) {
                        state = STATE_INIT;
                    }    
                    else    {
                        state = STATE_GET_CHAR;
                    }
                }
                break;

            case STATE_GET_CHAR:
                {
                    // Get char from text msg
                    char ch = ptrParms->textMsgBuffer[stringIndex++];

                    // Calculate index into font data array
                    charIndex = ch - ' ';

                    columnIndex = 0;

                    state = STATE_COLUMN_TEST;
                }
                break;

            case STATE_COLUMN_TEST:
                {
                    // Have all columns of char been displayed ?
                    if (columnIndex >= FONT_WIDTH) {
                        state = STATE_LENGTH_TEST;
                    }    
                    else {
                        state = STATE_GET_COLUMN;
                    }
                }
                break;

            case STATE_GET_COLUMN:
                {
                    // Get column data from font
                    uint8_t columnData = font6x8[charIndex][columnIndex++];

                    // Shift the font data to the left to make room for new column data
                    shiftFontData();

                    // Add columnData to last column of fontData
                    fontData[7] = columnData;

                    // Render font data on display
                    renderFontData(COLOR_WHITE, COLOR_BLACK);

                    // Calculate future time
                    delayTime = millis() + delayTimeMS;

                    state = STATE_DELAY;
                }
                break;

            case STATE_DELAY:
                {
                    if (millis() > delayTime) {
                        state = STATE_COLUMN_TEST;
                    }    
                }
                break;
            }
            return;
        }
    }
};

/*******************************************************************/
/***                        Patterns Mode                        ***/
/*******************************************************************/

// Instantiate all of the patterns
Rotation1    rotation1(&gParms);
RingChase2   ringChase2(&gParms);
RandomPTS    randomPTS(&gParms); 
RingChase    ringChase(&gParms);
Lines        lines(&gParms);
Chaser       chaser(&gParms);
Spiral1      spiral1(&gParms);
Fade         fade(&gParms);
Plasma       plasma(&gParms);
ColoredRings coloredRings(&gParms);
Rainbow1     rainbow1(&gParms);
Rainbow2     rainbow2(&gParms);

// Array of all patterns
BaseClass *patterns[] = {
    &plasma,
    &rotation1,
    &ringChase2,
    &randomPTS,
    &ringChase,
    &plasma,
    &lines,
    &chaser,
    &spiral1,
    &fade,
    &coloredRings,
    &rainbow1,
    &rainbow2
};

#define NUMBER_OF_PATTERNS (sizeof(patterns) / sizeof(BaseClass *))

// Pattern Mode
class PatternMode : 
public 
BaseClass {

    enum PMState {
        STATE_SELECT, STATE_SETUP, STATE_TIMEOUT
    };

    enum PMState state;

    BaseClass *currentPattern;
    long patternDisplayTime;

public:

    PatternMode(GLOBAL_PARAMETERS *_ptrParms) : 
    BaseClass(_ptrParms) {
    }

    void initStateMachine() {
        ptrParms->patternIndex = 0;

        // Set initial state
        state = STATE_SELECT;
    }

    void runStateMachine(boolean updated) {

        switch (state) {
        case STATE_SELECT:
            // Clear the matrix display in preparation for a new pattern
            clearFrameBuffer();
            
            delay(1000);

            if (ptrParms->sequentialPatterns) {
                // Sequential pattern mode
                currentPattern = patterns[ptrParms->patternIndex];
                ptrParms->patternIndex++;
                if (ptrParms->patternIndex >= NUMBER_OF_PATTERNS) {
                    ptrParms->patternIndex = 0;
                }
            }    
            else    {
                // Random pattern mode
                // Select random pattery
                int index = random(0, NUMBER_OF_PATTERNS);
                currentPattern = patterns[index];
                
                // Select a random speed as well
                ptrParms->speed = random(1, 10);
            }
            state = STATE_SETUP;
            break;

        case STATE_SETUP:
            // Run the init method on the pattern
            currentPattern->initStateMachine();

            // Calculate time to terminate pattern
            patternDisplayTime = millis() + PATTERN_DISPLAY_TIME_MS;

            state = STATE_TIMEOUT;
            break;

        case STATE_TIMEOUT:

            if (millis() > patternDisplayTime) {
                state = STATE_SELECT;
            }    
            else    {
                // Run the current pattern
                currentPattern->runStateMachine(updated);
            }
            break;
        }
    }
};   

#endif


























